// sb2html.mm -- convert SB instances into a set of interlinked 
//               files conforming to the HTML2 DTD using the text 
//               back-end.
//               This script will also generate a file "toc.html"
//               containing the generated table of contents for the
//               other files which are named "sec-N.html" where N is
//               the number of H1 document divisions in the source
//               document starting at 1.  By default, files will be
//               written into OUTDIR (cf .below)
//
// Copyright (C) 1995 MID/Information Logistics Group
//
// This file is part of the MetaMorphosis SGML conversion system.
// 
// $Id: sb2html.mm,v 1.6 1995/07/21 06:29:23 kfen Exp $

// C preprocessor command defining the target DTD

#define TARGETDTD "<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\">"

#ifndef OUTDIR
#define OUTDIR "/tmp/"
#endif

#define TRACE 1

///////////////////////// document element //////////////////////////

!TRANSFORMATION

// The rule for the document element SB simply prints out a message,
// then inserts the whole source tree in the target tree

sb	[]	 
	("running script sb2html.mm $Revision: 1.6 $ ...\n").echo(),
	r.contents;         

//////////////// navigational hyperlinks (on first page) ///////////

// The following set of transformation rules only apply in the 
// scope of the HEAD element 

!TRANSFORMATION (head)

// Create two new nodes POINTERS (also copy the contents of HEAD).
// These elements will end up as hypertext anchors both at the top and
// the bottom of each page allowing the user to navigate the HTML
// pages ("go to next, previous, last, first").

head	[]
	<|pointers|>node,r.contents,<|pointers|>node;

// For the node POINTERS (the page navigators) in the HEAD
// environment, create string content, two anchor elements, and more
// string content.  The attributes on those anchors are instantiated
// as follows: 
//  - on the first anchor, HREF-CNT is set to 1 and REL can
//    only be "next" 
//  - on the second one, HREF-CNT is set to the number
//    of H1 elements in the BODY of the document:
//    (count(source.child[?body].child[?h1])), while REL attribute value
//    can only be "last" (this being the last H1)

pointers
	[]
	"Go to the ",
	<|A, @href-cnt:=1,
	     @rel:="next"|>node{"NEXT, "},
	<|A, @href-cnt:= (count(source.child[?body].child[?h1])),
	     @rel:="last"|>node{"LAST "},
	"section";


//////////////////////// table of contents //////////////////////////

// The TOC element of the source document gets a whole subtree added.
// We start off by adding elements H1ENTRY that also get an
// attribute COUNT that simply counts the number of left siblings.
// The contents of any such new node is to be the TITLE element whose
// parent element in the source tree is an H1.

toc	 
  []	 
#if (TRACE)
     "creating table of contents ...\n".echo(),
#endif
    <|h1entry, @count:=count(r.parent.left[?h1]) + 1|>
          source.descendant[?title & parent[?h1]];

// We then continue populating the new H1ENTRYs: add a node TOCTIT,
// then create a node H2ENTRY under it similarly to the way H1 was treated.

h1entry	 
  []	 
    <|toctit|>node,
    <|h2entry, @count:=count(r.parent.left[?h2]) + 1|>
          r.parent.descendant[?title & parent[?h2]];

// Likewise, for the H2 and H3 elements.

h2entry	 
  []	 
    <|toctit|>node,
    <|h3entry, @count:=count(r.parent.left[?h3]) + 1|>
          r.parent.descendant[?title & parent[?h3]];

h3entry	 
  []	 
    <|toctit|>node,
    <|h4entry, @count:=count(r.parent.left[?h4]) + 1|>
          r.parent.descendant[?title & parent[?h4]];

// For the last level, simply create TOCTIT nodes

h4entry	 
  []	 
    <|toctit|>node;

// Populate the new TOCTIT nodes with nodes of type A (which will, of
// course, end up as HTML A elements).  These also get a couple of
// attributes: NAME-KEY is to be the MM-internal unique node id of the
// TITLE element copied above while HREF-CNT is the number of H1
// elements left of the current H1 under which the TITLE element lives
// (the use of this will become apparent below).

toctit	[]
	<|a, @name-key:=self.parent.r.key,
            @href-cnt:= (count(self.parent.r.ancestor[?h1].left) + 1)|>node;


// Here's an example of removing an element.

foreign-copy [] !replace;

// Now follows the set of annotation rules in the HEAD environment.
// In these and all following annotations, we use the built-in text
// output processor to add prefix and suffix strings to elements.
// Note that the whole expression after the keyword "_txt" must
// evaluate to a string.

!ANNOTATION (head)

// The former document title (ie, the TITLE element in the HEAD
// element) is to be the HTML element H1.

title	[]
	_txt := "prefix:<H1>#suffix:</H1>\n";

// The HEAD element turned table-of-contents element is to be a
// separate HTML page.  
// Note the use of the "file" statement (line 1) that directs the
// output to the named file.  Using this keyword means that the -T
// option on the command line should be "-" (for stdout).  The title
// of the page is to be the PCDATA contents of the TITLE element (line
// 8) plus the PCDATA contents of the SUBTITLE element (line 9).  This
// illustrates how easy it is to reorder the contents of elements in
// MM.  Note also that since we replaced the element FOREIGN-COPY in
// the target tree, we have to get its contents from the source tree
// (line 15).

head	[]	
          _txt := "file:" || OUTDIR || "toc.html" ||                    // 1
            "#prefix:" ||                                               // 2
            TARGETDTD ||                                                // 3
              "\n<!-- generated by MetaMorphosis on " ||                // 4
              date() ||                                                 // 5
              " -->\n" ||                                               // 6
              "<HTML>\n<HEAD>\n<TITLE>" ||                              // 7
              child[?title].data || " " ||                              // 8
              child[?subtitle].data ||                                  // 9
              "</TITLE>\n</HEAD>\n<BODY>\n" ||                          // 10
            "#suffix:" ||                                               // 11
             "<P>MetaMorphosis is Copyright &##169; 1995 " ||           // 12
             "MID/Information Logistics Group.</P>\n" ||                // 13
             "<P>" ||                                                   // 14
             r.descendant[?foreign-copy].data ||                        // 15
             "</P>\n</BODY></HTML>\n";                                  // 16



// The entries in the toc are wrapped in UL elements (unordered list)
// resulting in an indented list when the toc is rendered by the
// browser.

h1entry|h2entry|h3entry|h4entry
   []
   	_txt := "prefix:<UL>#suffix:</UL>\n";


// The actual toctitles are mapped to list items

toctit	[]	_txt := "prefix:<LI>#suffix:\n";


////////////////////////////// document structure /////////////////////

// The following set of transformations apply in the scope of the BODY
// element.

!TRANSFORMATION(body)

// Using the H1 elements we apply the same procedure of creating
// navigational hypertext links linking pages as above in the H1 rule
// (that created the first page in the HEAD environment).

h1	[] 
#if (TRACE)
      ("creating " || if(self[?h1], "chapter","appendix") || " " ||
       (count(self.left[?h1]) + 1) || "\n").echo(),
#endif  
	      <|pointers|>node, r.contents, <|pointers|>node;


// The contents of this POINTERS element is more complicated as we
// have to find out what the applicable navigational anchor elements
// will be.
// To do so we distinguish between the first H1, the last H1 and all
// other H1s in the document instance.  This is implemented in three
// contexts in which the rules for POINTERS apply:
// first: "[^(parent.left[?h1])]" meaning there is no H1 before
//        (ie. to the left) the current parent (which is itself an H1)
// last: "[^(parent.right[?h1])]" meaning there is no H1 after (ie. to
//       the right) the current parent
// others: "[]" meaning there's no context restriction to the rule
// application. 


pointers

//   first H1
	[^(parent.left[?h1])]
	"Go to the ",
	<|a, @href-cnt:= (count(self.parent.left[?h1]) + 2),
	     @rel:="next"|>node{"NEXT"},", ",
	<|a, @href-cnt := count(self.parent.parent.child[?h1]),
	     @rel:="last"|>node{"LAST"}," ",
	"section, or ",
	<|a, @rel:="TOC"|>node{"TABLE OF CONTENTS"},".";

//   last H1
	[^(parent.right[?h1])]
	"Go to the ",
	<|a, @href-cnt:=1,
	     @rel:="first"|>node{"FIRST"},", ",
	<|a, @href-cnt:= count(self.parent.r.left[?h1]),
	     @rel:="previous"|>node{"PREVIOUS"}," ",
	"section, or ",
	<|a, 
	     @rel:="TOC"|>node{"TABLE OF CONTENTS"},".";

//   all other H1
	[]
	"Go to the ",
	<|a, @href-cnt:=1,
	     @rel:="first"|>node{"FIRST"},", ",
	<|a, @href-cnt:= count(self.parent.left[?h1]),
	     @rel:="previous"|>node{"PREVIOUS"},", ",
	<|a, @href-cnt:= (count(self.parent.left[?h1]) + 2),
	     @rel:="next"|>node{"NEXT"},", ",
	<|a, @href-cnt:= count(self.parent.parent.child[?h1]),
	     @rel:="last"|>node{"LAST"}," ",
	"section, or ",
	<|a, 
	     @rel:="TOC"|>node{"TABLE OF CONTENTS"},".";


// All TITLEs that are direct children of the document sectioning
// elements H1--H4 of the document instance are to be anchors.  They
// receive an attribute NAME_KEY using the unique internal node id.
// Note that this node id was also used in the anchor attribute value
// in the TOCTIT rule above.  By doing so we create a two-way link
// from the toc entry anchor to the anchor in the HTML page and back.

title	[parent[?h1|?h2|?h3|?h4]]
	<|a, @name-key:=self.r.key|>node;
	[] r.contents;


// This rule deals with the cross-reference element XREF.  This rule
// also illustrates the use of variables in MM (*anchor in this case).
// Variables are always local to the rule in which they are declared.
// Note that instantiating the variable has the side effect of
// inserting the node it has as its value before the A element.  The
// variable determines the node that is the target anchor of the
// current XREF element denoted by the value of the REFID attribute.
// To find the anchor, we simply use the expression "@id ==
// self.@refid".  However, the context is further specified by
// requiring that the candidate not only have the matching ID value
// but also that it be a TITLE element whose parent must be an H1--H4
// element.  This has the effect that cross-references to other
// elements (such as TITLEs in TOPICs and ARTWORKs) will not be
// instantiated.  We assign also the two attributes NAME-KEY and
// HREF-CNT.  The former is again the unique node id of the TITLE
// element (ie, the value of the variable *anchor) while the HREF-CNT
// value is the number of H1 counted from the root of the source tree
// to the H1 that is an ancestor of the TITLE.

xref    [] 
       (*anchor :=source.descendant[?title & parent[?h1|?h2|?h3|?h4] & 
			            @id == self.@refid]),
       <|a, @name-key:=*anchor.key, 
	    @href-cnt:=*anchor.r.ancestor[?h1].position|>node;


// Junk all elements that in a formatting application would be the
// material for creating the index

ix1|ix2        []   !replace;


// The next rule is an example of string manipulation in MM.  The
// contents of the elements VERBATIM and CODE need to be searched for
// start tag open characters "<" and entity open characters "&".  Here
// we choose a staged approach as both VERBATIM and CODE have mixed
// content models: first, copy all PCDATA extents in temporary nodes
// MAKEENTS.  Then, in a rule for MAKEENTS, we use the parse procedure
// to replace the characters with the entities used in the HTML
// application. 

// This rule says: copy the contents of VERBATIM and CODE nodes such
// that nodes that have an ID (only element nodes have generic
// identifiers) receive their original ID while all other nodes (ie,
// PCDATA, entity and PI nodes) are to be named "MAKEENTS".

verbatim | code
   []
      <|id := id | "MAKEENTS" |> r.contents;

// Note the use of the @ symbol in the grammar: without it, parse would
// only perform the replacements but would throw away all characters
// that didn't match "<" and "&".

makeents []
    r.data.(parse(!begingram ("<" % "&lt;" | "&" % "&amp;" | @ )* . !endgram));



// Global annotations

!ANNOTATION

// The navigational anchors are wrapped in P elements.

pointers	[] _txt := "prefix:<P>#suffix:</P>\n<HR>\n";

///////////////////////// document division ////////////////////////////

// Instances of the SB DTD are to be split into several files.  Each
// file is to contain the contents of the document sectioning element
// H1.  The file name used in the "_txt" rule is synthesized by
// concatenating the string "sec-" with an integer that designates the
// number of left H1 siblings of the current H1.

h1	[]	
          _txt := "file:" || 
                    OUTDIR || "sec-" || (count(self.left) + 1) || ".html" ||
                  "#prefix:" ||
                    TARGETDTD ||
	            "\n<!-- generated by MetaMorphosis on " ||
	            date() || 
                    " -->\n" ||
                    "<HTML>\n<HEAD>\n<TITLE>" ||
                    child[?title].data ||
                    "</TITLE>\n</HEAD>\n<BODY>\n" ||
                  "#suffix:</BODY></HTML>\n";


// TITLE elements are mapped to a variety of HTML elements depending
// on their context.  Again, this is done by restricting the
// application of the rules by defining contexts.

// The first context is that of the XREF element that we copied above.
// Here we need to suppress the output of the contents of the TITLE
// element that we inserted under the XREF element (which was a side
// effect of declaring the local variable).  Note that the context
// rule refers to the local tree (ie the target tree).
// The context for the H1--H4 context in the -source- tree ensure that
// TITLE elements recieve the corresponding HTML sectioning element.
// Also, TITLEs of TOPICs will receive the H4 treatment.
// TITLEs whose parent is ARTWORK will be Paragraphs while we simply
// suppress the output of table titles.

title	[parent[?xref]] 
        _txt := "ignore";
        [r.parent[?h1]]
	_txt := "prefix:<H1>#suffix:</H1>\n";
	[r.parent[?h2]]
	_txt := "prefix:<H2>#suffix:</H2>\n";
	[r.parent[?h3]]
	_txt := "prefix:<H3>#suffix:</H3>\n";
	[r.parent[?h4]]
	_txt := "prefix:<H4>#suffix:</H4>\n";
	[r.parent[?topic]]
	_txt := "prefix:<H4>#suffix:</H4>\n";
	[r.parent[?artwork]]
	_txt := "prefix:<P>#suffix:</P>\n";
	[r.parent[?table]]
	_txt := "ignore";


// These were copied with the HEAD.  But we don't need their PCDATA
// contents. 

version | date | copy | author | subtitle

	[]
	_txt := "ignore";

//////////////////////// block elements ///////////////////////////////

// The P element cannot occur in all contexts.  In particular, P
// elements that lived in table cells (element ENTRY) receive null
// prefix and suffixes as we map tables to definition lists (see
// below).  All other Ps will become HTML Ps.


p	[r.parent[?entry]] _txt :=  "prefix:#suffix:";
        []                 _txt := "prefix:<P>#suffix:</P>\n";


// HTML has no footnotes.  Rather than creating separate pages for
// them, we simply wrap them with brackets and a descriptive string.

footnote []  _txt := "prefix: [NOTE: #suffix:] \n";

// These are rules for lists (BLIST = bulleted list; NLIST = numbered
// list; PLIST = prefix list; DLIST = definition list) and their
// constituent ITEMs.

blist
	[]	
	_txt := "prefix:<UL>#suffix:</UL>\n";

nlist
	[]	
	_txt := "prefix:<OL>#suffix:</OL>\n";

item	[parent[?pitem]]	
	_txt := "prefix:<DD>#suffix:</DD>\n";

	[]	
	_txt := "prefix:<LI>#suffix:\n";

plist|dlist
	[]	
	_txt := "prefix:<DL>#suffix:</DL><HR>\n";

prefix|term
	[]	
	_txt := "prefix:<DT>#suffix:</DT>\n";

def	[]	
	_txt := "prefix:<DD>#suffix:</DD>\n";


// Preformatted or verbatim text

verbatim
	[]	
	_txt := "prefix:<PRE>#suffix:</PRE>\n";


// Inline elements

code
	[]	
	_txt := "prefix:<CODE>#suffix:</CODE>";

emph1
	[]	
	_txt := "prefix:<B>#suffix:</B>";
emph2
	[]	
	_txt := "prefix:<I>#suffix:</I>";



//////////////////////////////// anchors ///////////////////////////////

// This set of transformations removes the IX (index entry) elements
// from the A elements (former TITLE elements)

!TRANSFORMATION

a	[parent[?toctit]]	ancestor:2.r.contents[^(?ix1)];
	[parent[?title]]	parent.r.contents[^(?ix1)];
	[]			r.contents[^(?ix1)];


// Now we need to output the anchor elements such that they point to
// the right target in the right file.

// Again, we distinguish a number of anchor elements depending on
// their ancestry.

// A elements that are children of POINTERS (the page navigators) and
// whose REL attribute has the value "TOC" are to be HTML A elements
// poiting to the file "toc.html"

// Other A children of POINTERS will have their HREF attributes
// instantiated to the number stored in their HREF-CNT attribut
// (recall that this simply denotes the number of H1 elements,
// ie. files, to the element).

// A elements whose grandmother are H1--H4 elements will point back to
// the corresponding entry in the table of contents thereby
// establishing a two-way link.  Note that the hash character in the
// HREF attribute value must be escaped (by another hash character)

// A elements whose parent is a TOCTIT (the toc entries) point to an
// anchor whose location in a file is determined by the HREF-CNT
// attribute value.

// All other A elements (the original cross-references in the
// document) make use of the same strategy.  However, they don't
// receive a NAME attribute (because we already have two-way links to
// some of their targets) and their content is taken from the subtrees
// that were inserted under the original XREF nodes.  The expression 
// "self.left:1.data" selects the PCDATA content of the first older
// sibling of the current element, in this case a TITLE element.

!ANNOTATION

a	[parent[?pointers & (self.@rel == "TOC")]]
		_txt := "prefix:<A  HREF=\"toc.html\" REL=\"" || 
                                @rel || 
                                "\">" ||
		        "#suffix:</A>";

a	[parent[?pointers]]
		_txt := "prefix:<A  HREF=\"sec-" || 
                                 @href-cnt  || ".html\"" ||
                                 " REL=\"" || @rel || "\">" ||
		        "#suffix:</A>";

	[ancestor:2[?h1|?h2|?h3|?h4]]
		_txt := "prefix:<A NAME=\"sec" || @name-key || "\"" || 
	                          " HREF=\"toc.html" || 
                                 "##" || "sec" ||
                                 @name-key ||
                                 "\"" ||">" ||
		        "#suffix:</A>"; 
	[parent[?toctit]]
		_txt := "prefix:<A NAME=\"sec" || @name-key || "\"" || 
	                         " HREF=\"sec-" || 
                                 @href-cnt || ".html" ||
                                 "##" || "sec" ||
                                 @name-key ||
                                 "\"" || ">" ||
		        "#suffix:</a>";
a       [] 
               _txt := "prefix:section \"<A HREF=\"sec-" || 
                               @href-cnt || ".html" ||
                               "##" || "sec" ||
                               @name-key ||
                               "\"" || ">" ||
		               self.left:1.data ||
		       "#suffix:</A>\"";


////////////////////////////////// graphics //////////////////////////////

// The following rule illustrates how to manipulate attribute values
// for the IMG element denoting graphics to be displayed by the
// browser.  Source instances contain the name of the graphics file in
// the FILE attribute.  In the MM manual these are EPS files ending in
// ".eps".  As Web browsers cannot display Postscript, they need to be
// renamed to *.gif.  The resultant tag also assumes that graphics
// live in a subdirectory named "graphics".

graphic
	[]
	_txt := "prefix:<IMG SRC=\"" ||
		 @file.(parse(!begingram (".eps" % ".gif" | @ )*. !endgram)) ||
                  "\">";

/////////////////////////////////// tables //////////////////////////////

// In this application we convert the CALS tables in the source
// instance to HTML definition lists DL (as HTML2 browsers don't
// support tables of any kind).

!ANNOTATION (table)

table	[]  
	_txt := "prefix:<DL>#suffix:</DL>\n"; 

thead	[]
	_txt := "ignore";

// The rule for ENTRYs (table cells) illustrates how easy it is to
// locate an element by its positions among its siblings using the
// keyword "position".  Those that are the first among their siblings
// will become definition terms DT, all others DDs.

entry	[position=1]
	_txt := "prefix:<DT>#suffix:</DT>\n";
	[]
	_txt := "prefix:<DD>#suffix:</DD>\n";

// sb2html.mm ends here

